package com.flyfish.framework.compiler.core;

import javax.tools.JavaCompiler;
import javax.tools.JavaFileObject;
import javax.tools.StandardJavaFileManager;
import javax.tools.ToolProvider;
import java.io.Closeable;
import java.io.IOException;
import java.util.Collections;
import java.util.Map;

/**
 * In-memory compile Java source code as String.
 *
 * @author michael
 */
public class JavaStringCompiler implements Closeable {

    private final JavaCompiler compiler;
    private final MemoryJavaFileManager manager;
    private final MemoryClassLoader classLoader;

    public JavaStringCompiler(MemoryClassLoader classLoader) {
        this.compiler = ToolProvider.getSystemJavaCompiler();
        StandardJavaFileManager stdManager = compiler.getStandardFileManager(null, null, null);
        this.manager = new MemoryJavaFileManager(stdManager, classLoader);
        this.classLoader = classLoader;
    }

    /**
     * Compile a Java source file in memory.
     *
     * @param fileName Java file name, e.g. "Test.java"
     * @param source   The source code as String.
     * @return The compiled results as Map that contains class name as key,
     * class binary as value.
     * @throws IOException If compile error.
     */
    public Map<String, byte[]> compile(String fileName, String source) {
        JavaFileObject javaFileObject = manager.makeStringSource(fileName, source);
        JavaCompiler.CompilationTask task = compiler.getTask(null, manager, null, null, null, Collections.singletonList(javaFileObject));
        Boolean result = task.call();
        if (result == null || !result) {
            throw new RuntimeException("Compilation failed.");
        }
        return manager.getClassBytes();
    }

    /**
     * Load class from compiled classes.
     *
     * @param name Full class name.
     * @return The Class instance.
     * @throws ClassNotFoundException If class not found.
     */
    public Class<?> loadClass(String name) throws ClassNotFoundException {
        return classLoader.loadClass(name);
    }

    @Override
    public void close() {
        try {
            if (null != manager) {
                manager.close();
            }
            if (null != classLoader) {
                classLoader.close();
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
}
