package com.flyfish.framework.repository.impl;

import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.domain.base.Qo;
import com.flyfish.framework.repository.DefaultRepository;
import com.flyfish.framework.utils.CopyUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.dao.IncorrectResultSizeDataAccessException;
import org.springframework.data.domain.Example;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.repository.query.MongoEntityInformation;
import org.springframework.data.mongodb.repository.support.SimpleMongoRepository;
import org.springframework.data.repository.support.PageableExecutionUtils;

import java.util.Collections;
import java.util.List;
import java.util.Optional;

/**
 * 查询模型支持
 *
 * @author wangyu
 * 基于repo的公共扩展
 */
public class DefaultRepositoryImpl<T extends Domain> extends SimpleMongoRepository<T, String>
        implements DefaultRepository<T> {

    private MongoOperations mongoOperations;
    private MongoEntityInformation<T, String> entityInformation;

    /**
     * Creates a new {@link SimpleMongoRepository} for the given {@link MongoEntityInformation} and {@link MongoTemplate}.
     *
     * @param metadata        must not be {@literal null}.
     * @param mongoOperations must not be {@literal null}.
     */
    public DefaultRepositoryImpl(MongoEntityInformation<T, String> metadata, MongoOperations mongoOperations) {
        super(metadata, mongoOperations);
        this.mongoOperations = mongoOperations;
        this.entityInformation = metadata;
    }

    /**
     * 通过名称查找一个
     *
     * @param name 名称
     * @return 结果
     */
    @Override
    public Optional<T> findByName(String name) {
        if (StringUtils.isNotBlank(name)) {
            return Optional.ofNullable(mongoOperations.findOne(Query.query(Criteria.where("name").is(name)),
                    entityInformation.getJavaType(), entityInformation.getCollectionName()));
        }
        return Optional.empty();
    }

    /**
     * Returns a single entity matching the given {@link Qo} or {@link Optional#empty()} if none was found.
     *
     * @param query must not be {@literal null}.
     * @return a single entity matching the given {@link Qo} or {@link Optional#empty()} if none was found.
     * @throws IncorrectResultSizeDataAccessException if the Qo yields more than one
     *                                                result.
     */
    @Override
    public Optional<T> findOne(Qo<T> query) {
        Query querying = getQuery(query);
        if (null != querying) {
            return Optional
                    .ofNullable(mongoOperations.findOne(querying,
                            entityInformation.getJavaType(),
                            entityInformation.getCollectionName()));
        }
        return Optional.empty();
    }

    /**
     * Returns all entities matching the given {@link Qo}. In case no match could be found an empty
     * {@link Iterable} is returned.
     *
     * @param query must not be {@literal null}.
     * @return all entities matching the given {@link Qo}.
     */
    @Override
    public Iterable<T> findAll(Qo<T> query) {
        Query querying = getQuery(query);
        if (null != querying) {
            return mongoOperations.find(querying,
                    entityInformation.getJavaType(),
                    entityInformation.getCollectionName());
        }
        return Collections.emptyList();
    }

    /**
     * Returns all entities matching the given {@link Qo} applying the given {@link Sort}. In case no match could
     * be found an empty {@link Iterable} is returned.
     *
     * @param query must not be {@literal null}.
     * @param sort  the {@link Sort} specification to sort the results by, may be {@link Sort#empty()}, must not be
     *              {@literal null}.
     * @return all entities matching the given {@link Qo}.
     * @since 1.10
     */
    @Override
    public Iterable<T> findAll(Qo<T> query, Sort sort) {
        return Collections.emptyList();
    }

    /**
     * Returns a {@link Page} of entities matching the given {@link Qo}. In case no match could be found, an empty
     * {@link Page} is returned.
     *
     * @param query    must not be {@literal null}.
     * @param pageable may be {@link Pageable#unpaged()}, must not be {@literal null}.
     * @return a {@link Page} of entities matching the given {@link Qo}.
     */
    @Override
    public Page<T> findAll(Qo<T> query, Pageable pageable) {
        Query querying = getQuery(query);
        if (null != querying) {
            querying.with(pageable);
            List<T> queryResult = mongoOperations.find(querying,
                    entityInformation.getJavaType(), entityInformation.getCollectionName());
            return PageableExecutionUtils.getPage(queryResult, pageable, () -> count(query));
        }
        return Page.empty();
    }

    /**
     * Returns the number of instances matching the given {@link Qo}.
     *
     * @param query the {@link Qo} to count instances for, must not be {@literal null}.
     * @return the number of instances matching the {@link Qo}.
     */
    @Override
    public long count(Qo<T> query) {
        Query q = getQuery(query);
        return this.mongoOperations.count(q, entityInformation.getJavaType(), entityInformation.getCollectionName());
    }

    /**
     * 通过特定键的集合查询
     *
     * @param key    键
     * @param values 集合
     * @return 结果
     */
    @Override
    public List<T> findAllByValues(String key, List<Object> values) {
        Criteria criteria = Criteria.where(key).in(values);
        Query query = new Query(criteria);
        return mongoOperations.find(query,
                entityInformation.getJavaType(),
                entityInformation.getCollectionName());
    }

    /**
     * Checks whether the data store contains elements that match the given {@link Qo}.
     *
     * @param query the {@link Qo} to use for the existence check, must not be {@literal null}.
     * @return {@literal true} if the data store contains elements that match the given {@link Qo}.
     */
    @Override
    public boolean exists(Qo<T> query) {
        return false;
    }

    /**
     * 从查询实体抽取内部查询信息
     *
     * @param qo 查询实体
     * @return 结果
     */
    private Query getQuery(Qo<T> qo) {
        Criteria criteria = null;
        if (null != qo.getCriteria()) {
            criteria = qo.getCriteria();
        } else if (null != qo.getExample()) {
            criteria = new Criteria().alike(qo.getExample());
        } else {
            Class<T> type = qo.pojoType();
            if (null != type && !Object.class.equals(type)) {
                try {
                    T pojo = CopyUtils.copyQueryProps(qo, qo.pojoType().newInstance());
                    criteria = new Criteria().alike(Example.of(pojo));
                } catch (InstantiationException | IllegalAccessException e) {
                    e.printStackTrace();
                }
            }
        }
        if (criteria != null) {
            // 针对删除状态全局筛选
            return new Query(Criteria.where("delete").ne(true).andOperator(criteria))
                    .with(qo.sorts());
        }
        return null;
    }

}
