package com.flyfish.framework.service;

import com.flyfish.framework.domain.AdminUserDetails;
import com.flyfish.framework.domain.base.IUser;
import com.flyfish.framework.domain.po.User;
import com.flyfish.framework.enums.UserStatus;
import com.flyfish.framework.utils.Assert;
import com.flyfish.framework.utils.CopyUtils;
import lombok.RequiredArgsConstructor;
import org.apache.commons.lang3.BooleanUtils;
import org.springframework.security.authentication.*;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.context.ReactiveSecurityContextHolder;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextImpl;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.security.crypto.password.NoOpPasswordEncoder;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.security.web.server.context.ServerSecurityContextRepository;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ServerWebExchange;
import reactor.core.publisher.Mono;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.util.HashMap;
import java.util.Map;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 * 基于mongo实现的用户详情
 *
 * @author wangyu
 */
@RequiredArgsConstructor
@Service
public class MongoUserDetailsServiceImpl implements MongoUserDetailsService {

    // 存储用户校验规则的map
    private static final Map<Function<User, Boolean>, Supplier<AuthenticationException>> checkMap;

    static {
        checkMap = new HashMap<>();
        // 初始化用户校验规则
        checkMap.put(user -> null != user.getEnable() && !user.getEnable() ||
                user.getStatus() == UserStatus.DISABLED, () -> new DisabledException("用户被禁用"));
        checkMap.put(user -> user.getStatus() == UserStatus.LOCKED,
                () -> new LockedException("账户已经锁定！请联系管理员修改密码！"));
    }

    private final ReactiveUserService userService;

    @Resource
    private ServerSecurityContextRepository contextRepository;

    private ReactiveAuthenticationManager authenticationManager;

    @PostConstruct
    @SuppressWarnings("deprecation")
    private void init() {
        UserDetailsRepositoryReactiveAuthenticationManager manager =
                new UserDetailsRepositoryReactiveAuthenticationManager(this);
        manager.setPasswordEncoder(NoOpPasswordEncoder.getInstance());
        authenticationManager = manager;
    }

    @Override
    public Mono<UserDetails> updatePassword(UserDetails userDetails, String s) {
        return userService.updateById(((IUser) userDetails).toUser()).map(this::mapToUserDetails);
    }

    @Override
    public Mono<UserDetails> findByUsername(String s) {
        return userService.findByUsername(s)
                .flatMap(this::validate)
                .map(this::mapToUserDetails)
                .switchIfEmpty(Mono.defer(() -> Mono.error(new UsernameNotFoundException("用户不存在！"))));
    }

    private UserDetails mapToUserDetails(User user) {
        AdminUserDetails userDetail = new AdminUserDetails();
        CopyUtils.copyProps(user, userDetail);
        return userDetail;
    }

    /**
     * 校验
     *
     * @param user 用户
     * @return 结果
     */
    private Mono<User> validate(User user) {
        return checkMap.entrySet().stream()
                .filter(entry -> entry.getKey().apply(user))
                .findFirst()
                .map(entry -> Mono.<User>error(entry.getValue().get()))
                .orElse(Mono.just(user));
    }

    /**
     * 通过用户名密码直接认证
     *
     * @param user 用户
     * @return 结果
     */
    @Override
    public Mono<Authentication> authenticate(UserDetails user, ServerWebExchange exchange) {
        return loadContext(user)
                .flatMap(securityContext -> contextRepository.save(exchange, securityContext)
                        .subscriberContext(ReactiveSecurityContextHolder.withSecurityContext(
                                Mono.just(securityContext)))
                        .then(Mono.just(securityContext)))
                .map(SecurityContext::getAuthentication)
                .doOnNext(authentication -> {
                    if (authentication.getPrincipal() instanceof IUser) {
                        IUser userDetail = (IUser) authentication.getPrincipal();
                        Assert.isTrue(BooleanUtils.isTrue(userDetail.getApp()), "用户没有移动端登录权限！");
                    }
                });
    }

    @Override
    public Mono<SecurityContext> loadContext(UserDetails user) {
        // 构建账号信息
        UsernamePasswordAuthenticationToken token = new UsernamePasswordAuthenticationToken(
                user.getUsername(), user.getPassword());
        // 登录先
        return authenticationManager.authenticate(token)
                .map(authentication -> {
                    SecurityContextImpl securityContext = new SecurityContextImpl();
                    securityContext.setAuthentication(authentication);
                    return securityContext;
                });
    }

    /**
     * 退出登录
     *
     * @param exchange 数据请求
     * @return 结果
     */
    @Override
    public Mono<Void> logout(ServerWebExchange exchange) {
        return ReactiveSecurityContextHolder.getContext()
                .switchIfEmpty(contextRepository.load(exchange))
                .flatMap(context -> contextRepository.save(exchange, context));

    }

    /**
     * 发生错误时的处理
     *
     * @param exchange 交换
     * @return 结果
     */
    @Override
    public Mono<User> error(ServerWebExchange exchange) {
        return exchange.getFormData().flatMap(data -> userService.findByUsername(data.getFirst("username")))
                .flatMap(user -> {
                    User updating = new User();
                    updating.setId(user.getId());
                    updating.setErrorCount(user.getErrorCount() + 1);
                    if (updating.getErrorCount() >= 5) {
                        updating.setStatus(UserStatus.LOCKED);
                    }
                    return userService.updateSelectiveById(updating);
                });
    }
}
